----------------------------------------------------------------------------------
-- Company:  INSA-Toulouse
-- Engineer: Paul Faure
-- 
-- Create Date: 18.04.2021 21:19:41 
-- Module Name: Etage3_Calcul - Structural
-- Project Name: Processeur sécurisé
-- Target Devices: Basys 3 ARTIX7
-- Tool Versions: Vivado 2016.4
--
-- Description: Etage 3 du processeur
--     - Gestion de l'ALU
-- 
-- Dependencies: 
--    - ALU
--    - LC
--    - MUX
--
-- Additional Comments: Etage assynchrone
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity Etage3_Calcul is
    Generic ( Nb_bits : Natural; -- Taille d'un mot binaire
              Instruction_bus_size : Natural; -- Nombre de bits du bus d'instruction (Taille d'un code instruction)
              Bits_Controle_LC : STD_LOGIC_VECTOR; -- Vecteur de bit controlant le Link Controler (cf LC.vhd)
              Bits_Controle_MUX : STD_LOGIC_VECTOR); -- Vecteur de bit controlant le multiplexeur (cf MUX.vhd)
    Port ( RST : in STD_LOGIC; -- Reset
           IN_A : in STD_LOGIC_VECTOR (Nb_bits - 1 downto 0); -- Entrée de l'opérande A
           IN_B : in STD_LOGIC_VECTOR (Nb_bits - 1 downto 0); -- Entrée de l'opérande B
           IN_C : in STD_LOGIC_VECTOR (Nb_bits - 1 downto 0); -- Entrée de l'opérande C
           IN_Instruction : in STD_LOGIC_VECTOR (Instruction_bus_size - 1 downto 0); -- Entrée de l'instruction
           OUT_A : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0); -- Sortie de l'opérande A
           OUT_B : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0); -- Sortie de l'opérande B
           OUT_Instruction : out STD_LOGIC_VECTOR (Instruction_bus_size - 1 downto 0); -- Sortie de l'instruction
           N : out STD_LOGIC; -- Flag Negative
           O : out STD_LOGIC; -- Flag Overload
           Z : out STD_LOGIC; -- Flag Zero
           C : out STD_LOGIC);-- Flag Carry
end Etage3_Calcul;

architecture Structural of Etage3_Calcul is
    component ALU is
    Generic (Nb_bits : Natural);
    Port ( A : in STD_LOGIC_VECTOR (Nb_bits-1 downto 0);
           B : in STD_LOGIC_VECTOR (Nb_bits-1 downto 0);
           OP : in STD_LOGIC_VECTOR (2 downto 0);
           S : out STD_LOGIC_VECTOR (Nb_bits-1 downto 0);
           N : out STD_LOGIC;
           O : out STD_LOGIC;
           Z : out STD_LOGIC;
           C : out STD_LOGIC);
    end component;
    
    component LC is
        Generic (Instruction_Vector_Size : Natural;
                Command_size : Natural;
                Bits_Controle : STD_LOGIC_VECTOR);
        Port ( Instruction : in STD_LOGIC_VECTOR (Instruction_Vector_Size - 1 downto 0);
               Commande : out STD_LOGIC_VECTOR (Command_size - 1 downto 0));
    end component;
    
    component MUX is
        Generic (Nb_bits : Natural;
                Instruction_Vector_Size : Natural;
                Bits_Controle : STD_LOGIC_VECTOR);
        Port ( Instruction : in STD_LOGIC_VECTOR (Instruction_Vector_Size - 1 downto 0);
               IN1 : in STD_LOGIC_VECTOR (Nb_bits - 1 downto 0);
               IN2 : in STD_LOGIC_VECTOR (Nb_bits - 1 downto 0);
               OUTPUT : out STD_LOGIC_VECTOR (Nb_bits - 1 downto 0));
    end component;
    
    -- Sortie du Link Controleur commandant l'ALU
    signal OP_ALU : STD_LOGIC_VECTOR (2 downto 0) := (others => '0');
    
    -- Sortie de l'ALU, a passer au multiplexeur
    signal Sortie_ALU : STD_LOGIC_VECTOR (Nb_bits - 1 downto 0) := (others => '0');
    
    
    -- signaux internes
    signal intern_OUT_B : STD_LOGIC_VECTOR (Nb_bits - 1 downto 0) := (others => '0');
    signal intern_N : STD_LOGIC := '0';
    signal intern_O : STD_LOGIC := '0';
    signal intern_Z : STD_LOGIC := '0';
    signal intern_C : STD_LOGIC := '0';
    
    
begin
    instance_LC : LC
    generic map (Instruction_Vector_Size => Instruction_bus_size,
                 Command_size => 3,
                 Bits_Controle => Bits_Controle_LC)
    port map (   Instruction => IN_Instruction,
                 Commande => OP_ALU);
                 
    instance_MUX : MUX
    generic map (Nb_bits => Nb_bits,
                 Instruction_Vector_Size => Instruction_bus_size,
                 Bits_Controle => Bits_Controle_MUX)
    port map (   Instruction => IN_Instruction,
                 IN1 => IN_B,
                 IN2 => Sortie_ALU,
                 OUTPUT => intern_OUT_B);

    instance_ALU : ALU
    generic map (Nb_bits => Nb_bits)
    port map    (A => IN_B,
                 B => IN_C,
                 OP => OP_ALU,
                 S => Sortie_ALU,
                 N => intern_N,
                 O => intern_O,
                 Z => intern_Z,
                 C => intern_C);
    
    OUT_A <= (others => '0') when RST = '0' else
             IN_A;
    OUT_B <= (others => '0') when RST = '0' else
             intern_OUT_B;
    OUT_Instruction <= (others => '0') when RST = '0' else
             IN_Instruction;
    N <= '0' when RST = '0' else
             intern_N;
    O <= '0' when RST = '0' else
             intern_O;
    Z <= '0' when RST = '0' else
             intern_Z;
    C <= '0' when RST = '0' else
             intern_C;
end Structural;
